@@ -19,9 +19,11 @@ import sys
 import traceback
 import urlparse
 import zipfile
+import xml.dom.minidom
+import mimetypes
 
 PROG = os.path.basename(os.path.splitext(__file__)[0])
-VERSION = '8.6.4'
+VERSION = '8.6.5'
 
 # AsciiDoc global configuration file directory.
 # NOTE: CONF_DIR is "fixed up" by Makefile -- don't rename or change syntax.
@@ -418,16 +420,27 @@ class A2X(AttrDict):
             for r in open(self.resource_manifest):
                 self.resources.append(r.strip())
         for r in self.resources:
-            if os.path.isfile(r):
-                self.resource_files.append(r)
+            r = os.path.expanduser(r)
+            r = os.path.expandvars(r)
+            if r.endswith(('/','\\')):
+                if  os.path.isdir(r):
+                    self.resource_dirs.append(r)
+                else:
+                    die('missing resource directory: %s' % r)
             elif os.path.isdir(r):
                 self.resource_dirs.append(r)
+            elif r.startswith('.') and '=' in r:
+                ext, mimetype = r.split('=')
+                mimetypes.add_type(mimetype, ext)
             else:
-                die('missing resource: %s' % r)
-        # Lastly search among images and stylesheets distributed with asciidoc.
+                self.resource_files.append(r)
         for p in (os.path.dirname(self.asciidoc), CONF_DIR):
             for d in ('images','stylesheets'):
-                self.resource_dirs.append(os.path.join(p,d))
+                d = os.path.join(p,d)
+                if os.path.isdir(d):
+                    self.resource_dirs.append(d)
+        verbose('resource files: %s' % self.resource_files)
+        verbose('resource directories: %s' % self.resource_dirs)
         if not self.doctype and self.format == 'manpage':
             self.doctype = 'manpage'
         if self.doctype:
@@ -523,12 +536,10 @@ class A2X(AttrDict):
         '''
         Search html_files for images and CSS resource URIs (html_files can be a
         list of file names or a single file name).
-        If the URIs are relative then copy them from the src_dir to the
-        dst_dir.
+        Copy them from the src_dir to the dst_dir.
         If not found in src_dir then recursively search all specified
-        resources for the file name.
-        Optional additional resources URIs can be passed in the resources list.
-        Does not copy absolute URIs.
+        resource directories.
+        Optional additional resources files can be passed in the resources list.
         '''
         resources = resources[:]
         resources += find_resources(html_files, 'link', 'href',
@@ -538,24 +549,32 @@ class A2X(AttrDict):
         resources = list(set(resources))    # Drop duplicates.
         resources.sort()
         for f in resources:
-            f = os.path.normpath(f)
-            if os.path.isabs(f):
-                if not os.path.isfile(f):
-                    warning('missing resource: %s' % f)
-                continue
-            src = os.path.join(src_dir, f)
-            dst = os.path.join(dst_dir, f)
+            if '=' in f:
+                src, dst = f.split('=')
+                if not dst:
+                    dst = src
+            else:
+                src = dst = f
+            src = os.path.normpath(src)
+            dst = os.path.normpath(dst)
+            if os.path.isabs(dst):
+                die('absolute resource file name: %s' % dst)
+            if dst.startswith(os.pardir):
+                die('resource file outside destination directory: %s' % dst)
+            src = os.path.join(src_dir, src)
+            dst = os.path.join(dst_dir, dst)
             if not os.path.isfile(src):
                 for d in self.resource_dirs:
-                    src = find_files(d, os.path.basename(f))
-                    if src:
-                        src = src[0]
+                    d = os.path.join(src_dir, d)
+                    found = find_files(d, os.path.basename(src))
+                    if found:
+                        src = found[0]
                         break
                 else:
                     if not os.path.isfile(dst):
-                        warning('missing resource: %s' % f)
-                    continue    # Continues outer for loop.
-            # Arrive here if relative resource file has been found.
+                        die('missing resource: %s' % src)
+                    continue
+            # Arrive here if resource file has been found.
             if os.path.normpath(src) != os.path.normpath(dst):
                 dstdir = os.path.dirname(dst)
                 shell_makedirs(dstdir)
@@ -648,6 +667,45 @@ class A2X(AttrDict):
         src_dir = os.path.dirname(self.asciidoc_file)
         self.copy_resources(html_files, src_dir, dst_dir)
 
+    def update_epub_manifest(self, opf_file):
+        '''
+        Scan the OEBPS directory for any files that have not been registered in
+        the OPF manifest then add them to the manifest.
+        '''
+        opf_dir = os.path.dirname(opf_file)
+        resource_files = []
+        for (p,dirs,files) in os.walk(os.path.dirname(opf_file)):
+            for f in files:
+                f = os.path.join(p,f)
+                if os.path.isfile(f):
+                    assert f.startswith(opf_dir)
+                    f = '.' + f[len(opf_dir):]
+                    f = os.path.normpath(f)
+                    if f not in ['content.opf']:
+                        resource_files.append(f)
+        opf = xml.dom.minidom.parseString(open(opf_file).read())
+        manifest_files = []
+        manifest = opf.getElementsByTagName('manifest')[0]
+        for el in manifest.getElementsByTagName('item'):
+            f = el.getAttribute('href')
+            f = os.path.normpath(f)
+            manifest_files.append(f)
+        count = 0
+        for f in resource_files:
+            if f not in manifest_files:
+                count += 1
+                verbose('adding to manifest: %s' % f)
+                item = opf.createElement('item')
+                item.setAttribute('href', f.replace(os.path.sep, '/'))
+                item.setAttribute('id', 'a2x-%d' % count)
+                mimetype = mimetypes.guess_type(f)[0]
+                if mimetype is None:
+                    die('unknown mimetype: %s' % f)
+                item.setAttribute('media-type', mimetype)
+                manifest.appendChild(item)
+        if count > 0:
+            open(opf_file, 'w').write(opf.toxml())
+
     def to_epub(self):
         self.to_docbook()
         xsl_file = self.xsl_stylesheet()
@@ -658,16 +716,17 @@ class A2X(AttrDict):
         shell_makedirs(build_dir)
         # Create content.
         exec_xsltproc(xsl_file, docbook_file, build_dir, self.xsltproc_opts)
-        # Copy OPF file resources.
+        # Copy resources referenced in the OPF and resources referenced by the
+        # generated HTML (in theory DocBook XSL should ensure they are
+        # identical but this is not always the case).
         src_dir = os.path.dirname(self.asciidoc_file)
         dst_dir = os.path.join(build_dir, 'OEBPS')
-        # Get the resources from OPF instead of HTML content to get round this
-        # bug: https://sourceforge.net/tracker/?func=detail&aid=2854080&group_id=21935&atid=373747
-        opf = os.path.join(dst_dir, 'content.opf')
-        resources = find_resources(opf, 'item', 'href')
-#        html_files = find_files(dst_dir, '*.html')
-#        self.copy_resources(html_files, src_dir, dst_dir)
-        self.copy_resources([], src_dir, dst_dir, resources)
+        opf_file = os.path.join(dst_dir, 'content.opf')
+        opf_resources = find_resources(opf_file, 'item', 'href')
+        html_files = find_files(dst_dir, '*.html')
+        self.copy_resources(html_files, src_dir, dst_dir, opf_resources)
+        # Register any unregistered resources.
+        self.update_epub_manifest(opf_file)
         # Build epub archive.
         cwd = os.getcwd()
         shell_cd(build_dir)
@@ -797,7 +856,7 @@ if __name__ == '__main__':
     parser.add_option('--stylesheet',
         action='store', dest='stylesheet', default=None,
         metavar='STYLESHEET',
-        help='target HTML CSS stylesheet file name')
+        help='HTML CSS stylesheet file name')
     #DEPRECATED
     parser.add_option('--safe',
         action='store_true', dest='safe', default=False,
@@ -813,7 +872,7 @@ if __name__ == '__main__':
         metavar='FOP_OPTS', help='options for FOP pdf generation')
     parser.add_option('--xsltproc-opts',
         action='append', dest='xsltproc_opts', default=[],
-        metavar='XSLTPROC_OPTS', help='options for FOP pdf generation')
+        metavar='XSLTPROC_OPTS', help='xsltproc options for XSL stylesheets')
     parser.add_option('--xsl-file',
         action='store', dest='xsl_file', metavar='XSL_FILE',
         help='custom XSL stylesheet')
